'' /////////////////////////////////////////////////////////////////////////////
'' Graphics Renderer for Donkey Kong Remake
'' (C) Steve Waddicor
'' //////////////////////////////////////////////////////////////////////////////

CON
  PARAM_COUNT   = 7
  NUMBER_OF_STRINGS = 32

OBJ
  tv    : "sw_dk_tv_drv_022.spin"

VAR
  long  cog

VAR
    ' Renderer parameter block
    ' Each renderer copies these values to it's own cog_memory cache
    ' when it's got a little spare time, several times per frame
    ' so these values can be altered on the fly.
    long    rend_cog_number        'each cog gets a different number, starting with zero     write-only
    long    rend_cog_count         'total number of cogs                                     write-only
    long    rend_scanline_req      'a pointer to the request to put a line in the scan buffer   write-only
    long    rend_scanline_buffer   'hub memory where renderer puts scanline when requested write-only
    long    rend_text_location_table
    long    rend_text_color_table
    long    rend_text_string_table
    'End of renderer parameter block

    ' TV parameter block
    long    tv_scanline_request
    long    tv_scanline_buffer
    long    tv_status
    ' End of TV parameter block

    long    string_locations[32]
    long    end
    'string_colors
    '0 = invisible
    byte    string_colors[32]
    'string_ptrs
    '0 = this string unused.  OK to reuse.
    word    string_ptrs[32]

PUB Start(aCogs,aTvStatusPtr) | i

    repeat i from 0 to NUMBER_OF_STRINGS-1
        string_locations[i]:=$FFFF_FFFF    
        string_colors[i] := 0
        string_ptrs[i] := 0
    end:=$FFFF_FFFF
    string_locations[0]:=$3800_6000
    string_colors[0] := $8e
    string_ptrs[0] := @hi
    
    rend_cog_count := aCogs
    rend_scanline_req := @scanline_req
    rend_scanline_buffer := @hub_scanline_buffer
    rend_text_location_table := @string_locations
    rend_text_color_table := @string_colors
    rend_text_string_table := @string_ptrs

    repeat i from 0 to aCogs-1
        rend_cog_number := i
        cognew(@entry,@rend_cog_number)
        'Wait to allow this cog to get the cog number param from the block before changing it.
        repeat 1000

    tv_scanline_request := @scanline_req
    tv_scanline_buffer := @hub_scanline_buffer
    tv.start(@tv_scanline_request)
    'return address of tv_status to caller
    long[aTvStatusPtr]:=@tv_status

PUB NewString(aId,aX,aY,aString,aChroma,aLuma):rHandle
    {{
    aId:            An identifier for referring to the string.  This must be in
                    the range 0 =< aId < NUMBER_OF_STRINGS.  Strings with
                    higher IDs display in front of strings with lower IDs.
                    
    word aX,aY:     Location of bottom left of string.  Coordinates have their
                    origin at the top left of the screen and y-axis increases
                    down the screen.  Coodinates are in fixed point form, with
                    4 bits for the fractional part.  Thus:
                        $0000,$00F0 will display a string at the top left of the screen.
                        $0000,$0100 will display one pixel lower down the screen.
                        $0000,$0DF0 will display on the bottom line of the screen.

    string aString: The string to display

    byte    aColor: The color to display the text in.  This is a hi-color value.
                    The special value of 0 will make the text invisible.
    }}
    if aId=>NUMBER_OF_STRINGS
        return -1
    string_locations[aId]:=(aX<<16)+aY
    string_ptrs[aId]:=aString               
    string_colors[aId]:=((aChroma&$0f)<<4) + aLuma&$3 + 11

PUB DeleteString(aId)
    string_ptrs[aId]~

PUB DeltaStringPos(aId,aDeltaX,aDeltaY)
    string_locations[aId]+=aDeltaX<<16+aDeltaY
    
PUB DeltaStringColor(aId,aDeltaChroma,aDeltaLuma) | chroma, luma
    chroma:=string_colors[aId]>>4
    luma  :=(string_colors[aId] - 3)&%11
    chroma+=aDeltaChroma
    luma+= aDeltaLuma
    luma #>= 0
    luma <#= 3
    string_colors[aId]:=(chroma&$0f)<<4 + luma + 11

PUB MoveStringTo(aId,aX,aY)
    string_locations[aId]:=(aX<<16)+aY
    
DAT
hi      byte  "Fantasia!",0    
DAT

'******************************
'* Assembly language renderer *
'******************************
                        org
'
'
' Entry
'
entry
                        rdlong  cog_rend_cog_number,PAR
                        or      dira,#1
                        MOV     OUTA,#1
                        mov     current_line,cog_rend_cog_number  'start rendering from the top
do_line
                        ' Load parameters
                        mov     r0,PAR
                        movd    :load_dest,#renderer_params+1
                        mov     count,#param_count-1
:load                   add     r0,#4
                        '       ___
:load_dest              rdlong  0-0,r0
                        '       ^^^ iterates through renderer_params array.
                        add     :load_dest,inc_dest
                        djnz    count,#:load

'@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
                        movd    :clear_line,#cog_scanline_buffer
                        mov     count,#64
:clear_line             mov     0-0,black
                        add     :clear_line,inc_dest
                        djnz    count,#:clear_line

                        mov     hub_location_tab_ptr,cog_rend_text_location_table
                        mov     hub_color_tab_ptr,cog_rend_text_color_table
                        mov     hub_string_tab_ptr,cog_rend_text_string_table
:string_loop            rdlong  location,hub_location_tab_ptr
                        cmp     location,end_of_location_array wz
              if_e      jmp     #:end_string_table
              
                        mov     y,location
                        shr     y,#4
                        and     y,FFF
                        mov     char_line,current_line
                        sub     char_line,y wz,wc
              if_b      jmp     #:next_string

                        cmp     char_line,#32 wz,wc
              if_ae     jmp     #:next_string
                        shl     char_line,#2
              
                        rdbyte  color,hub_color_tab_ptr wz
              if_z      jmp     #:next_string

                        mov     pix_ptr,#cog_scanline_buffer
                        mov     x,location
                        shr     x,#22     
                        add     pix_ptr,x

                        rdword  string_ptr,hub_string_tab_ptr
:char_loop              rdbyte  char_ptr,string_ptr wz
              if_z      jmp     #:next_string
                        mov     offset,char_ptr
                        
                        shr     char_ptr,#1
                        shl     char_ptr,#7
                        add     char_ptr,char_set_start
                        add     char_ptr,char_line

                        rdlong  char,char_ptr wz

                        shr     offset,#1 wc
              if_c      shr     char,#1

                        mov     nybble_count,#4

:next_nybble            cmps    pix_ptr,#cog_scanline_buffer wz,wc
              if_b      jmp     #:skip_off_screen
                        cmps    pix_ptr,#end_cog_scanline_buffer wz,wc
              if_ae     jmp     #:skip_off_screen

                        movs    :load1,pix_ptr
                        movd    :save1,pix_ptr
:load1                  mov     pixels,0-0

                        mov     pixel_count,#4
:next_pixel             shr     char,#2 wc
              if_c      and     pixels,ffffff00
              if_c      or      pixels,color
                        ror     pixels,#8
                        djnz    pixel_count,#:next_pixel

:save1                  mov     0-0,pixels

:skip_off_screen        add     pix_ptr,#1 
                        djnz    nybble_count,#:next_nybble
                        
                        add     string_ptr,#1
                        jmp     #:char_loop
                        
:next_string            add     hub_string_tab_ptr,#2
                        add     hub_color_tab_ptr,#1
                        add     hub_location_tab_ptr,#4
                        call    #checktv
              if_b      jmp     #:string_loop
                        
:end_string_table
                        
'@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

                        ' We've finished rendering the line and we didn't have a hissy fit panic.
                        ' We need to wait here for the line we just prepared to be requested by the TV driver
wait_for_the_request
                        call    #checktv
              if_b      jmp     #wait_for_the_request

                        ' Our line has been requested so copy it to hub scan line buffer
start_tv_copy           movs    :nextcopy, #cog_scanline_buffer
                        mov     r0, cog_rend_scanline_buffer
                        mov     count, #64              ' length of visible scanline buffer in longs
:nextcopy               mov     r1,cog_scanline_buffer
                        add     :nextcopy, #1
                        wrlong  r1, r0
                        add     r0, #4
                        djnz    count, #:nextcopy

                        ' We've handed over the scanline to the TV driver, so do another.
allocate_next_job       add     current_line,cog_rend_cog_count 'we only do every Nth line
                        ' Degrade gracefully when there's too much work to do.
                        call    #checktv
              if_ae     jmp     #allocate_next_job
                       ' Wrap around.
                        cmpsub  current_line,#224   'cmpsub will subtract 224 if current_line is greater than 224.
                        jmp     #do_line


'-------------------------------------------------------------------
checktv                 ' Tf we're currently rendering a low numbered line when a high one is being requested
                        ' then that's OK because we wrapped around.



                        rdlong  scanline_req,cog_rend_scanline_req
                        cmp     scanline_req,#210 wc,wz
              if_be     jmp     #:checkit
                        cmp     current_line,#7 wc,wz
              if_be     cmp     scanline_req, current_line wz
                        'b = normal
                        'z = doit
                        'a = err
              if_be     jmp     #checktv_ret

:checkit                cmp     scanline_req, current_line wz, wc    'set carry if request is less than current.
                        'b = normal
                        'z = doit
                        'a = err

checktv_ret             ret
'-------------------------------------------------------------------

' Initialized data
black                   long    $02_02_02_02
end_of_location_array   long    $FFFF_FFFF
char_set_start          long    $8000
ffffff00                long    $FFFFFF00
fff                     long    $FFF
inc_dest                long    1 << 9 << 0

hub_location_tab_ptr    long    0
hub_color_tab_ptr       long    0
hub_string_tab_ptr      long    0
char_ptr                long    0
char                    long    0
location                long    0
x                       long    0
y                       long    0
color                   long    0
pix_ptr                 long    0
pixels                  long    0
string_ptr              long    0
offset                  long    0
char_line               long    0
nybble_count            long    0
pixel_count             long    0

hub_scanline_buffer
colorbar0               long    $0b_0b_0b_0b
colorbar1               long    $0c_0c_0c_0c
colorbar2               long    $0d_0d_0d_0d
colorbar3               long    $0e_0e_0e_0e
colorbar4               long    $1b_0b_1b_1b
colorbar5               long    $1c_0c_1c_1c
colorbar6               long    $1d_0d_1d_1d
colorbar7               long    $1e_0e_1e_1e
colorbar8               long    $2b_0b_2b_2b
colorbar9               long    $2c_2c_2c_2c
colorbar10              long    $2d_2d_2d_2d
colorbar11              long    $2e_2e_2e_2e
colorbar12              long    $3b_3b_3b_3b
colorbar13              long    $3c_3c_3c_3c
colorbar14              long    $3d_3d_3d_3d
colorbar15              long    $3e_3e_3e_3e
colorbar16              long    $4b_4b_4b_4b
colorbar17              long    $4c_4c_4c_4c
colorbar18              long    $4d_4d_4d_4d
colorbar19              long    $4e_4e_4e_4e
colorbar20              long    $5b_5b_5b_5b
colorbar21              long    $5c_5c_5c_5c
colorbar22              long    $5d_5d_5d_5d
colorbar23              long    $5e_5e_5e_5e
colorbar24              long    $6b_6b_6b_6b
colorbar25              long    $6c_6c_6c_6c
colorbar26              long    $6d_6d_6d_6d
colorbar27              long    $6e_6e_6e_6e
colorbar28              long    $7b_7b_7b_7b
colorbar29              long    $7c_7c_7c_7c
colorbar30              long    $7d_7d_7d_7d
colorbar31              long    $7e_7e_7e_7e
colorbar32              long    $8b_8b_8b_8b
colorbar33              long    $8c_8c_8c_8c
colorbar34              long    $8d_8d_8d_8d
colorbar35              long    $8e_8e_8e_8e
colorbar36              long    $9b_9b_9b_9b
colorbar37              long    $9c_9c_9c_9c
colorbar38              long    $9d_9d_9d_9d
colorbar39              long    $9e_9e_9e_9e
colorbar40              long    $ab_ab_ab_ab
colorbar41              long    $ac_ac_ac_ac
colorbar42              long    $ad_ad_ad_ad
colorbar43              long    $ae_ae_ae_ae
colorbar44              long    $bb_bb_bb_bb
colorbar45              long    $bc_bc_bc_bc
colorbar46              long    $bd_bd_bd_bd
colorbar47              long    $be_be_be_be
colorbar48              long    $cb_cb_cb_cb
colorbar49              long    $cc_cc_cc_cc
colorbar50              long    $cd_cd_cd_cd
colorbar51              long    $ce_ce_ce_ce
colorbar52              long    $db_db_db_db
colorbar53              long    $dc_dc_dc_dc
colorbar54              long    $dd_dd_dd_dd
colorbar55              long    $de_de_de_de
colorbar56              long    $eb_eb_eb_eb
colorbar57              long    $ec_ec_ec_ec
colorbar58              long    $ed_ed_ed_ed
colorbar59              long    $ee_ee_ee_ee
colorbar60              long    $fb_fb_fb_fb
colorbar61              long    $fc_fc_fc_fc
colorbar62              long    $fd_fd_fd_fd
colorbar63              long    $fe_fe_fe_fe
' End of scanline buffer
scanline_req            long    0

DAT 'Uninitiaised data - always keep it at the end of file to avoid nasty gotcha with RES.

renderer_params  
cog_rend_cog_number             res 1 'each cog gets a different number, starting with zero     write-only
cog_rend_cog_count              res 1 'total number of cogs                                     write-only
cog_rend_scanline_req           res 1 'a pointer to the request to put a line in the scan buffer   write-only
cog_rend_scanline_buffer        res 1 'hub memory where renderer puts scanline when requested write-only
cog_rend_text_location_table    res 1
cog_rend_text_color_table       res 1
cog_rend_text_string_table      res 1
    
' Cog scanline buffer = 256 pixels, 8 bits per pixel = 2048 bits = 256 bytes = 64 longs
cog_scanline_buffer     res     64
end_cog_scanline_buffer

current_line            res     1
r0                      res     1
r1                      res     1
count                   res     1

                        fit
